from __future__ import division
import numpy as np
import matplotlib.pyplot as plt
from matplotlib         import rc, rcParams
from annulus_analytical import compute_eigenvalues, compute_eigenvector
from annulus_numerical  import eigen_decomposition
from filter             import filter_drp15



# Font/plot setup
rc('text', usetex=True)
rc('font', family='serif')
rcParams.update({'figure.autolayout': True})
fig1 = plt.figure()
fig2 = plt.figure()
fig3 = plt.figure()
fig1.set_facecolor('White')
fig2.set_facecolor('White')
fig3.set_facecolor('White')
ax1 = fig1.add_subplot(111)
ax2 = fig2.add_subplot(111)
ax3 = fig3.add_subplot(111)



# Define mean state
gam = 1.4
rho = 1.2212179
u   = 0.
v   = 0.
w   = 103.2586448
p   = 103341.6659



# Define circumferential and temporal wavenumber
m     = 2
omega = 3441.9548



# Define geometry and discretization
res = 50
ri = 0.25
ro = 1.0
dr = (ro-ri)/(res-1)
nfields = 5
r = np.linspace(ri,ro,res)
dof = res*nfields



# Compute eigen-decomposition
eigenvalues, eigenvectors = eigen_decomposition(m,omega,ri,ro,rho,u,v,w,p,gam)



# Separate eigenvectors into primitive variables
rho_eigenvectors = eigenvectors[0*res:1*res,:]
u_eigenvectors   = eigenvectors[1*res:2*res,:]
v_eigenvectors   = eigenvectors[2*res:3*res,:]
w_eigenvectors   = eigenvectors[3*res:4*res,:]
p_eigenvectors   = eigenvectors[4*res:5*res,:]



# Create separate copy so we can compare before/after modes
pr_eigenvectors   = np.copy(np.real(p_eigenvectors))
pr_eigenvectors_f = np.copy(pr_eigenvectors)



# Call filter for each eigenvector
for i in range(pr_eigenvectors.shape[1]):
    pr_eigenvectors_f[:,i] = filter_drp15(pr_eigenvectors[:,i])

# Select low-wavenumber modes based on the ratio of their vector norms
amp   = np.zeros(pr_eigenvectors.shape[1])
amp_f = np.zeros(pr_eigenvectors.shape[1])
for i in range(pr_eigenvectors.shape[1]):
    for j in range(len(r)-1):
        amp[i]   = amp[i]   + 0.5*(r[j+1]-r[j])*(r[j]*pr_eigenvectors[  j,i]*pr_eigenvectors[  j,i]  +  r[j+1]*pr_eigenvectors[  j+1,i]*pr_eigenvectors[  j+1,i])
        amp_f[i] = amp_f[i] + 0.5*(r[j+1]-r[j])*(r[j]*pr_eigenvectors_f[j,i]*pr_eigenvectors_f[j,i]  +  r[j+1]*pr_eigenvectors_f[j+1,i]*pr_eigenvectors_f[j+1,i])

# Compute selection criteria
alpha = amp_f/amp

# Plot vectors that satisfy criteria
for i in range(len(alpha)):
    if (alpha[i] < 0.00001):     # res = 50
        if (np.imag(eigenvalues[i]) > 0.):
            h_up, = ax1.plot(np.real(eigenvalues[i]),np.imag(eigenvalues[i]),   'b^',markersize=6, label='Acc. Down')
        elif (np.imag(eigenvalues[i]) < 0.):
            h_down, = ax1.plot(np.real(eigenvalues[i]),np.imag(eigenvalues[i]), 'bs',markersize=6, label='Acc. Up'  )
        ax2.plot(pr_eigenvectors[:,i]/max(abs(pr_eigenvectors[:,i])), r, '-')






# Good to compare against aiaaj_iftn_2009 Fig 13
#m  = 10
#Ri = 0.52
#Ro = 1.0

mach = 0.3
omega = 10
sigma = ri





# Get eigenvalues for each radial mode
nrmodes = 10
alpha = np.zeros(nrmodes)
alpha = compute_eigenvalues(m,nrmodes,ri,ro)

# Compute axial wave numbers
k = np.zeros(2*len(alpha),dtype=np.complex)
for i in range(len(alpha)):
    a = mach*mach - 1.
    b = 2.*omega*mach
    c = omega*omega - alpha[i]*alpha[i]
    roots = np.roots([a,b,c])
    k[2*i]   = roots[0]
    k[2*i+1] = roots[1]
h_analytical, = ax1.plot(k.real,k.imag, 'o', markerfacecolor='None',markersize=10,label='Analytical')




# Compute radius and percent span location
r_array = np.linspace(ri,ro,100)
p_span  = np.zeros(100)
for i in range(len(r_array)):
    p_span[i] = 100.*(r_array[i]-ri)/(ro-ri)

# Loop through the radial modes 
for rmode in range(nrmodes):
    eigenvector = compute_eigenvector(r_array,sigma,m,alpha[rmode])
    ax3.plot(eigenvector,p_span)
    




ax1.legend(handles=[h_analytical,h_up,h_down],numpoints=1)
ax1.set_xlabel('$Re(k_z)$')
ax1.set_ylabel('$Im(k_z)$')
#ax1.set_xlim((-15.,10.))
#ax1.set_ylim((0.6,1.0))
ax3.set_ylim((0,100))
ax3.set_xlim((-1.1,1.1))
ax3.set_xlabel('Radial mode amplitude', weight='bold', fontsize=12, labelpad=10)
ax3.set_ylabel('$\%$ Span',             weight='bold', fontsize=12, labelpad=10)


#fig.set_size_inches(5,8)
#fig.savefig('radial_modes_1d.png', bbox_inches='tight')
plt.show()






# -*- coding: utf-8 -*-
r""" 
Numerical
---------

This module provides a functionality for computing the numerical eigenvalue/eigenvector
decomposition of a uniform axial flow in an annular cylindrical duct. The decomposition
is based on a normal mode analysis of the three-dimensional linearized Euler equations,
which yields an eigensystem that is discretized and solved numerically.


Theory:
~~~~~~~



Filtering:
~~~~~~~~~~



Example:
~~~~~~~~

::
    
    eigenvalues, eigenvectors, r = noisyduck.annulus.numerical.decomposition(omega,m,ri,ro,rho,u,v,w,p,gam)

"""
import numpy as np
import scipy
import noisyduck.filter






def decomposition(omega,m,r,rho,u,v,w,p,gam,filter='None',alpha=0.00001):
    r""" Compute the numerical eigen-decomposition of the three-dimensional linearized
    Euler equations for a cylindrical annulus.

    Args:
        omega (float): temporal frequency.
        m (int): circumferential wavenumber.
        r (float): array of equally-spaced radius locations for the discretization, including end points.
        rho (float): mean density.
        u (float): mean radial velocity.
        v (float): mean tangential velocity.
        w (float): mean axial velocity.
        p (float): mean pressure.
        gam (float): ratio of specific heats.
        filter (string, optional): Optional filter for eigenmodes. allowable values = ['None', 'acoustic']
        alpha (float, optional): Criteria governing filtering acoustic modes. 

    Returns:
        (eigenvalues, eigenvectors, r): a tuple containing an array of eigenvalues, an array 
        of eigenvectors evaluated at radial locations, and an array of those radial locations.

    Note:
        The eigenvectors being returned include each field :math:`[\rho,u,v,w,p]`. The primitive
        variables can be extracted into their own eigenvectors by copying out those entries from 
        the returned eigenvectors as:

        ::

            res = len(r)
            rho_eigenvectors = eigenvectors[0*res:1*res,:]
            u_eigenvectors   = eigenvectors[1*res:2*res,:]
            v_eigenvectors   = eigenvectors[2*res:3*res,:]
            w_eigenvectors   = eigenvectors[3*res:4*res,:]
            p_eigenvectors   = eigenvectors[4*res:5*res,:]

    """
    res = len(r)
    ri = np.min(r)
    ro = np.max(r)


    # Construct eigensystem
    M, N = construct_numerical_eigensystem(omega,m,r,rho,u,v,w,p,gam)

    # Solve Generalized Eigenvalue Problem for complex, nonhermitian system
    eigenvalues, eigenvectors = scipy.linalg.eig(M,N,right=True,overwrite_a=True,overwrite_b=True)
    
    # Add radial velocity end points back where they were removed due to boundary conditions
    eigenvectors = np.insert(eigenvectors, [res]    , [0.] ,axis=0)
    eigenvectors = np.insert(eigenvectors, [2*res-1], [0.] ,axis=0)

    # Potential filtering
    if (filter == 'acoustic'):
        eigenvalues, eigenvectors = noisyduck.filter.physical(eigenvalues,eigenvectors,r,alpha_cutoff=alpha,filters=filter)

    return eigenvalues, eigenvectors


    


def construct_numerical_eigensystem(omega,m,r,rho,u,v,w,p,gam):
    """ Constructs the numerical representation of the eigenvalue problem associated
    with the three-dimensional linearized euler equations subjected to a normal mode
    analysis.

    NOTE: a small imaginary part is added to the temporal frequency to facilitate 
    determining the propagation direction of eigenmodes based on the sign of the 
    imaginary part of their eigenvalue.

    Args:
        omega (float): temporal frequency.
        m (int): circumferential wavenumber.
        r (float): array of equally-spaced radius locations for the discretization, including end points.
        rho (float): mean density.
        u (float): mean radial velocity.
        v (float): mean tangential velocity.
        w (float): mean axial velocity.
        p (float): mean pressure.
        gam (float): ratio of specific heats.

    Returns:
        (M, N, r): left-hand side of generalized eigenvalue problem, right-hand side 
        of generalized eigenvalue problem, and radial coordinates of the discretization.
    """

    # Define real/imag parts for temporal frequency
    romega = omega
    iomega = -10.e-5*romega

    # Define geometry and discretization
    res = len(r)
    ri = np.min(r)
    ro = np.max(r)
    dr = (ro-ri)/(res-1)
    nfields = 5
    dof = res*nfields


    # Allocate storage
    M  = np.zeros([dof,dof], dtype=np.complex)
    N  = np.zeros([dof,dof], dtype=np.complex)
    vl = np.zeros([dof,dof], dtype=np.complex)
    vr = np.zeros([dof,dof], dtype=np.complex)
    
    # Submatrices for discretization
    stencil     = np.zeros([res,res])
    identity    = np.zeros([res,res])
    ridentity   = np.zeros([res,res])
    zero_matrix = np.zeros([res,res])

    # Construct fourth-order finite difference stencil
    stencil[0,0:5] = [-25., 48., -36, 16., -3.]
    stencil[1,0:5] = [-3., -10., 18., -6., 1.]
    stencil[res-2, res-5:res] = [-1., 6., -18., 10., 3.]
    stencil[res-1, res-5:res] = [3., -16., 36., -48., 25.]
    for i in range(2,res-2):
        stencil[i,i-2] =  1.
        stencil[i,i-1] = -8.
        stencil[i,i+1] =  8.
        stencil[i,i+2] = -1.
    # Scale by dr
    stencil = (1./(12.*dr))*stencil
    
    
    # Construct identity matrix for source terms
    for i in range(res):
        identity[i,i] = 1.
    
    
    # Construct identity scaled by 1/r
    for i in range(res):
        ridentity[i,i] = 1./r[i]
    

    # Block 1,1
    irow = 1
    icol = 1
    irow_start = 0 + res*(irow-1)
    icol_start = 0 + res*(icol-1)
    M[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( M[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        + 1j*identity*romega        # temporal source(real)
                                                                        -    identity*iomega        # temporal source(imag)
                                                                        +    stencil*u              # radial derivative
                                                                        +    ridentity*u            # radial source
                                                                        + 1j*ridentity*float(m)*v ) # circumferential source

    N[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( N[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        + 1j*identity*w )           # axial source

    # Block 2,2
    irow = 2
    icol = 2
    irow_start = 0 + res*(irow-1)
    icol_start = 0 + res*(icol-1)
    M[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( M[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        + 1j*identity*romega        # temporal source(real)
                                                                        -    identity*iomega        # temporal source(imag)
                                                                        +    stencil*u              # radial derivative
                                                                        +    ridentity*u            # radial source
                                                                        + 1j*ridentity*float(m)*v ) # circumferential source

    N[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( N[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        + 1j*identity*w )           # axial source


    # Block 3,3
    irow = 3
    icol = 3
    irow_start = 0 + res*(irow-1)
    icol_start = 0 + res*(icol-1)
    M[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( M[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        + 1j*identity*romega        # temporal source(real)
                                                                        -    identity*iomega        # temporal source(imag)
                                                                        +    stencil*u              # radial derivative
                                                                        +    ridentity*u            # radial source
                                                                        + 1j*ridentity*float(m)*v   # circumferential source
                                                                        +    ridentity*u )          # equation/coordinate source

    N[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( N[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        + 1j*identity*w )           # axial source



    # Block 4,4
    irow = 4
    icol = 4
    irow_start = 0 + res*(irow-1)
    icol_start = 0 + res*(icol-1)
    M[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( M[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        + 1j*identity*romega        # temporal source(real)
                                                                        -    identity*iomega        # temporal source(imag)
                                                                        +    stencil*u              # radial derivative
                                                                        +    ridentity*u            # radial source
                                                                        + 1j*ridentity*float(m)*v ) # circumferential source

    N[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( N[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        + 1j*identity*w )           # axial source


    # Block 5,5
    irow = 5
    icol = 5
    irow_start = 0 + res*(irow-1)
    icol_start = 0 + res*(icol-1)
    M[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( M[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        + 1j*identity*romega        # temporal source(real)
                                                                        -    identity*iomega        # temporal source(imag)
                                                                        +    stencil*u              # radial derivative
                                                                        +    ridentity*u            # radial source
                                                                        + 1j*ridentity*float(m)*v   # circumferential source
                                                                        +    ridentity*(gam-1.)*u ) # equation/coordinate source

    N[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( N[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        + 1j*identity*w )           # axial source


    # Block 2,1
    irow = 2
    icol = 1
    irow_start = 0 + res*(irow-1)
    icol_start = 0 + res*(icol-1)
    M[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( M[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        - ridentity*v*v/rho )   # equation/coordinate source

    # Block 3,1
    irow = 3
    icol = 1
    irow_start = 0 + res*(irow-1)
    icol_start = 0 + res*(icol-1)
    M[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( M[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        + ridentity*u*v/rho )   # equation/coordinate source


    # Block 1,2
    irow = 1
    icol = 2
    irow_start = 0 + res*(irow-1)
    icol_start = 0 + res*(icol-1)
    M[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( M[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        + stencil*rho           # radial derivative
                                                                        + ridentity*rho )       # radial source

    # Block 3,2
    irow = 3
    icol = 2
    irow_start = 0 + res*(irow-1)
    icol_start = 0 + res*(icol-1)
    M[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( M[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        + ridentity*v )         # equation/coordinate source

    # Block 5,2
    irow = 5
    icol = 2
    irow_start = 0 + res*(irow-1)
    icol_start = 0 + res*(icol-1)
    M[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( M[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        + stencil*gam*p                 # radial derivative
                                                                        + ridentity*gam*p               # radial source
                                                                        - ridentity*(gam-1.)*rho*v*v )  # equation/coordinate source

    # Block 1,3
    irow = 1
    icol = 3
    irow_start = 0 + res*(irow-1)
    icol_start = 0 + res*(icol-1)
    M[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( M[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        + 1j*ridentity*rho*float(m) )   # circumferential source

    # Block 2,3
    irow = 2
    icol = 3
    irow_start = 0 + res*(irow-1)
    icol_start = 0 + res*(icol-1)
    M[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( M[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        - ridentity*2.*v )              # equation/coordinate source

    # Block 5,3
    irow = 5
    icol = 3
    irow_start = 0 + res*(irow-1)
    icol_start = 0 + res*(icol-1)
    M[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( M[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        + 1j*ridentity*gam*p*float(m)       # circumferential source
                                                                        +    ridentity*(gam-1.)*rho*u*v )   # equation/coordinate source

    # Block 1,4
    irow = 1
    icol = 4
    irow_start = 0 + res*(irow-1)
    icol_start = 0 + res*(icol-1)
    N[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( N[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        + 1j*identity*rho )                # axial source

    # Block 5,4
    irow = 5
    icol = 4
    irow_start = 0 + res*(irow-1)
    icol_start = 0 + res*(icol-1)
    N[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( N[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        + 1j*identity*gam*p )              # axial source

    # Block 2,5
    irow = 2
    icol = 5
    irow_start = 0 + res*(irow-1)
    icol_start = 0 + res*(icol-1)
    M[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( M[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        + stencil/rho                   # radial derivative
                                                                        + ridentity/rho                 # radial source
                                                                        - ridentity/rho )               # equation/coordinate source
    # Block 3,5
    irow = 3
    icol = 5
    irow_start = 0 + res*(irow-1)
    icol_start = 0 + res*(icol-1)
    M[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( M[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        + 1j*ridentity*float(m)/rho )   # circumferential source

    # Block 4,5
    irow = 4
    icol = 5
    irow_start = 0 + res*(irow-1)
    icol_start = 0 + res*(icol-1)
    N[irow_start:irow_start + (res), icol_start:icol_start + (res)] = ( N[irow_start:irow_start+(res),icol_start:icol_start+(res)] 
                                                                        + 1j*identity/rho )             # axial source



    # Remove rows/columns due to solid wall boundary condition: no velocity normal to boundary, assumes 
    # here wall normal vector is aligned with radial coordinate.
    M = np.delete(M, (res,2*res-1), axis=0)
    M = np.delete(M, (res,2*res-1), axis=1)
    
    N = np.delete(N, (res,2*res-1), axis=0)
    N = np.delete(N, (res,2*res-1), axis=1)
    
    
    # Move N to right-hand side
    N = -N


    return M, N








#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages
from setuptools.command.build_ext import build_ext as _build_ext



class build_ext(_build_ext):
    def finalize_options(self):
        _build_ext.finalize_options(self)
        # Prevent numpy from thinking it is still in its setup process:
        __builtins__.__NUMPY_SETUP__ = False
        import numpy
        self.include_dirs.append(numpy.get_include())


with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    'Click>=6.0',
    # TODO: put package requirements here
    'numpy',
    'scipy',
    'matplotlib'
]

setup_requirements = [
    'pytest-runner',
    # TODO(nwukie): put setup requirements (distutils extensions, etc.) here
]

test_requirements = [
    'pytest',
    # TODO: put package test requirements here
]

setup(
    name='noisyduck',
    version='0.1.4',
    description="A python tool for computing eigenmode decompositions of duct flows.",
    long_description=readme + '\n\n' + history,
    author="Nathan Wukie",
    author_email='nathan.wukie@gmail.com',
    url='https://github.com/nwukie/noisyduck',
    #packages=find_packages(include=['noisyduck']),
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'noisyduck=noisyduck.cli:main'
        ]
    },
    include_package_data=True,
    install_requires=requirements,
    license="BSD license",
    zip_safe=False,
    keywords='noisyduck',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
    test_suite='tests',
    tests_require=test_requirements,
    setup_requires=setup_requirements,
)
